Attribute VB_Name = "modInterface"
Option Explicit

' ----==== Const ====----
Private Const S_OK As Long = &H0&
Private Const CLSCTX_INPROC As Long = &H1&
Private Const CC_STDCALL As Long = &H4&
Private Const IID_Release As Long = &H8&
Private Const ERR_BASE As Long = &H9EDE&

' ----==== Interface Error Codes ====----
Public Enum InterfaceErrCodes
    ecd_None                    ' no error
    ecd_NoInterfacePtr          ' no interface pointer
    ecd_InitInterface           ' could not create interface
    ecd_ReleaseInterface        ' could not release interface
    ecd_OleConvertCLSID         ' could not convert classid
    ecd_OleConvertIID           ' could not convert interface id
    ecd_OleInvoke               ' could not invoke interface function
    ecd_InvalidCall             ' invalid function call
End Enum

' ----==== Hlt div. Interface Daten ====----
Public Type InterfaceData
    pInterface As Long          ' Interface-Pointer
    pObject As Long             ' Pointer zur Klasse
    boolRaiseErrors As Boolean  ' Fehler auslsen?
End Type

' ----==== Kernel32 API-Deklarationen ====----
Private Declare Sub RtlMoveMemory Lib "kernel32" ( _
                    ByRef Destination As Any, _
                    ByRef Source As Any, _
                    ByVal Length As Long)

' ----==== OleAut32 API-Deklarationen ====----
Private Declare Function DispCallFunc Lib "OleAut32" ( _
                    ByVal pvInstance As Long, _
                    ByVal oVft As Long, _
                    ByVal cc As Long, _
                    ByVal vtReturn As VbVarType, _
                    ByVal cActuals As Long, _
                    ByVal prgvt As Any, _
                    ByVal prgpvarg As Any, _
                    ByRef pvargResult As Variant) As Long

' ----==== Ole32 API-Deklarationen ====----
Private Declare Function CLSIDFromString Lib "ole32" ( _
                         ByVal lpsz As Long, _
                         ByVal pCLSID As Long) As Long

Private Declare Function CoCreateInstance Lib "ole32" ( _
                         ByVal rclsid As Long, _
                         ByVal pUnkOuter As Long, _
                         ByVal dwClsContext As Long, _
                         ByVal riid As Long, _
                         ByRef ppv As Long) As Long

' ----==== Init Interface ====----
Public Function InitInterface(ByRef tInterface As InterfaceData, ByVal sCLSID As _
    String, ByVal sIID As String) As Boolean

    Dim bIID() As Byte
    Dim bCLSID() As Byte

    ' IID zu ByteArray
    If OleConvert(sIID, bIID) = False Then

        InitInterface = False

        ' Fehler auslsen
        Call InterfaceError(tInterface, ecd_OleConvertIID)

        Exit Function

    End If

    ' CLSID zu ByteArray
    If OleConvert(sCLSID, bCLSID) = False Then

        InitInterface = False

        ' Fehler auslsen
        Call InterfaceError(tInterface, ecd_OleConvertCLSID)

        Exit Function

    End If
        
    ' Interfcae aus CLSID und IID erstellen
    If CoCreateInstance(VarPtr(bCLSID(0)), 0&, CLSCTX_INPROC, VarPtr(bIID(0)), _
        tInterface.pInterface) <> S_OK Then

        InitInterface = False

        ' Fehler auslsen
        Call InterfaceError(tInterface, ecd_InitInterface)

    Else

        InitInterface = True

    End If

End Function

' ----==== Release Interface ====----
Public Function ReleaseInterface(ByRef tInterface As InterfaceData) As Boolean

    Dim vRet As Variant
    Dim bRet As Boolean

    bRet = False

    ' ist ein Pointer auf ein Interfac vorhanden
    If tInterface.pInterface Then

        ' Funktion Release des Interfaces aufrufen
        If DispCallFunc(tInterface.pInterface, IID_Release, CC_STDCALL, vbLong, _
            0&, 0&, 0&, vRet) = S_OK Then

            bRet = True

        Else

            ' Fehler auslsen
            Call InterfaceError(tInterface, ecd_ReleaseInterface)

        End If

    End If

    ReleaseInterface = bRet

End Function

' ----==== Call Interface Function ====----
Public Function OleInvoke(ByRef tInterface As InterfaceData, ByVal lCmd As Long, _
    ParamArray aParam()) As Variant
    Dim lCount As Long
    Dim lItem As Long
    Dim vRet As Variant
    Dim vParam As Variant
    Dim olePtr(10) As Long
    Dim oleTyp(10) As Integer

    ' wenn kein Pointer auf ein Interface vorhanden ist
    If tInterface.pInterface = 0& Then

        ' Fehler auslsen
        Call InterfaceError(tInterface, ecd_NoInterfacePtr)

    Else

        ' nur wenn zum Aufruf der Interface-Funktion auch
        ' Parameter vorhanden sind.
        If UBound(aParam) >= 0 Then

            ' ParamArray nach Variant
            vParam = aParam

            ' ist der Variant ein Array
            If IsArray(vParam) Then vParam = vParam(0)

            ' Anzahl der Parameter
            lCount = UBound(vParam)

            ' alle Parameter durchlaufen
            For lItem = 0 To lCount

                ' Typ des Parameter
                oleTyp(lItem) = VarType(vParam(lItem))

                ' Pointer auf den Parameter
                olePtr(lItem) = VarPtr(vParam(lItem))

            Next

        End If

        ' Funktion des Interfaces aufrufen
        If DispCallFunc(tInterface.pInterface, lCmd * 4, CC_STDCALL, vbLong, _
            lItem, VarPtr(oleTyp(0)), VarPtr(olePtr(0)), vRet) <> S_OK Then

            ' Fehler auslsen
            Call InterfaceError(tInterface, ecd_InvalidCall)
                
        End If

    End If

    OleInvoke = vRet

End Function

' ----==== Interface Error ====----
Public Function InterfaceError(ByRef tInterface As InterfaceData, Optional ByVal _
    eErrCode As InterfaceErrCodes = -1) As Boolean

    Dim oDmy As Object
    Dim oObj As Object
    Dim sErrText As String

    ' ist ein Fehlercode <> ecd_None vorhanden
    If eErrCode <> ecd_None Then

        ' Fehlercodes auswerten
        Select Case eErrCode

        Case ecd_NoInterfacePtr
            sErrText = "no interface pointer"

        Case ecd_InitInterface
            sErrText = "could not create interface"

        Case ecd_ReleaseInterface
            sErrText = "could not release interface"

        Case ecd_OleConvertCLSID
            sErrText = "could not convert classid"

        Case ecd_OleConvertIID
            sErrText = "could not convert interface id"

        Case ecd_OleInvoke
            sErrText = "could not invoke interface function"

        Case ecd_InvalidCall
            sErrText = "invalid function call"

        End Select

        With tInterface

            ' soll ein Fehler ausgelst werden
            If .boolRaiseErrors = True Then

                ' ist ein Pointer auf eine Klasse vorhanden
                If .pObject <> 0& Then

                    ' Objekt vom Pointer
                    Call RtlMoveMemory(oDmy, .pObject, 4)

                    ' Objekt speichern
                    Set oObj = oDmy

                    ' Objekt lschen
                    Call RtlMoveMemory(oDmy, 0&, 4)

                    ' Fehlertext generieren
                    sErrText = "Error in " & TypeName(oObj) & ": " & sErrText

                    ' Sub x_RaiseError in der entsprechenden Klasse aufrufen
                    Call oObj.x_RaiseError(ERR_BASE + eErrCode, sErrText)

                End If

            End If

        End With

    End If

End Function

' ----==== IID/CLSID zu ByteArray ====----
Private Function OleConvert(ByVal sIidClsid As String, ByRef bIidClsid() As Byte) _
    As Boolean

    Dim bRet As Boolean

    bRet = False

    ReDim bIidClsid(15)

    If CLSIDFromString(StrPtr(sIidClsid), VarPtr(bIidClsid(0))) = S_OK Then

        bRet = True

    End If

    OleConvert = bRet

End Function
